"""Envarify implementation."""

from __future__ import annotations

import inspect
import os
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Protocol, Type

from .errors import AnnotationError, MissingEnvVarsError
from .parse import EnvVarParser, SupportedType, get_parser


@dataclass(frozen=True)
class EnvVar:
    """Environment variable specification.

    Arguments:
        name: Environment variable name.
            If not specified, attribute name will be used instead
        default: Default value. Makes environment variable optional
            i.e. error will not be raised if it is not set
        parse: Custom parse function
        delimiter: Delimiter string. Only applicable for parsing sequences
    """

    name: str | None = None
    default: str | None = None
    parse: EnvVarParser | None = None
    delimiter: str = ","


class BaseConfig:
    """Base config class."""

    def __init__(self, **kwargs) -> None:
        """Initialize this object."""
        for key, value in kwargs.items():
            if key not in self._annotations():
                raise TypeError("Unexpected keyword argument '{}'".format(key))
            setattr(self, key, value)

    def __repr__(self) -> str:
        """Return representation."""
        return "{name}({attributes})".format(
            name=self.__class__.__name__,
            attributes=", ".join(
                "{key}={value!r}".format(key=key, value=getattr(self, key))
                for key in self._annotations()
            ),
        )

    @classmethod
    def fromenv(cls) -> BaseConfig:
        """Initialize this object from environment variables.

        Returns:
            BaseConfig

        Raises:
            AnnotationError - if subclass' annotation are not valid
            MissingEnvVarsError - if required environment variables are not set
            ValueError - if environment variables values cannot be parsed into specified type
        """
        cls.validate()
        return cls(**{v.attr: v.value() for v in cls._sources()})

    @classmethod
    def validate(cls) -> None:
        """Run validations required to initialize from env vars."""
        not_annotated = [key for key in cls._properties() if key not in cls._annotations()]
        if not_annotated:
            raise AnnotationError(
                "Missing type annotatations in the following properties: "
                + ", ".join(not_annotated)
            )

        # collect from this config class
        missing_envvars = [e.name for e in cls._envvars() if not e.has_value()]

        # collect from child config classes
        for base_class in cls._base_classes():
            try:
                base_class.config_type.validate()
            except MissingEnvVarsError as e:
                missing_envvars.extend(e.env_vars)

        if missing_envvars:
            raise MissingEnvVarsError(missing_envvars)

    @classmethod
    @lru_cache
    def _properties(cls) -> dict[str, Any]:
        """Get dictionary containg class properties."""
        return {k: v for k, v in cls.__dict__.items() if not k.startswith("__")}

    @classmethod
    def _annotations(cls) -> dict[str, Type]:
        """Get dictionary containg class properties."""
        if not hasattr(cls, "__annotations__"):
            raise AnnotationError("Subclass must have annotated properties")

        return cls.__annotations__

    @classmethod
    @lru_cache
    def _envvars(cls) -> list[_EnvVarSource]:
        """Yield over all environment variables."""
        return [source for source in cls._sources() if isinstance(source, _EnvVarSource)]

    @classmethod
    @lru_cache
    def _base_classes(cls) -> list[_BaseConfigSource]:
        """Yield over all base config sources."""
        return [source for source in cls._sources() if isinstance(source, _BaseConfigSource)]

    @classmethod
    @lru_cache
    def _sources(cls) -> list[_ValueSource]:
        """Yield over all value sources."""
        sources: list[_ValueSource] = []

        for key, type_ in cls._annotations().items():
            spec: EnvVar | None = cls._properties().get(key)

            if inspect.isclass(type_) and issubclass(type_, BaseConfig):
                sources.append(_BaseConfigSource(attr=key, config_type=type_))
            elif isinstance(spec, EnvVar):
                sources.append(
                    _EnvVarSource(
                        attr=key,
                        name=spec.name or key,
                        default=spec.default,
                        parse=spec.parse or get_parser(type_, spec),
                    )
                )
            elif spec is None:
                sources.append(
                    _EnvVarSource(
                        attr=key,
                        name=key,
                        parse=get_parser(
                            type_, EnvVar()  # EnvVar() is passed because it has default values
                        ),
                    )
                )
        return sources


class _ValueSource(Protocol):

    attr: str

    def value(self) -> SupportedType | BaseConfig | None: ...


@dataclass(frozen=True)
class _BaseConfigSource(_ValueSource):
    """Base config source."""

    attr: str
    config_type: Type[BaseConfig]

    def value(self) -> BaseConfig:
        return self.config_type.fromenv()


@dataclass(frozen=True)
class _EnvVarSource(_ValueSource):
    """Environment variable representation."""

    attr: str
    name: str
    parse: EnvVarParser
    default: str | None = None

    def exists(self) -> bool:
        """Check if this environment variable exists."""
        return self.name in os.environ

    def has_value(self) -> bool:
        """Check if this environment variable has value."""
        return self.default is not None or self.exists()

    def value(self) -> SupportedType | None:
        """Check if this environment variable has value."""
        if self.has_value():
            value = os.environ.get(self.name, default=self.default)
            return self.parse(value)  # type: ignore
        return None



"""Host package errors."""

from __future__ import annotations


class EnvarifyError(Exception):
    """Base module exception."""


class MissingEnvVarsError(EnvarifyError):
    """Missing environment variables error."""

    def __init__(self, env_vars: list[str]) -> None:
        """Initialize error."""
        super().__init__(", ".join(env_vars))
        self.env_vars = env_vars


class AnnotationError(EnvarifyError):
    """Annotation error."""


class UnsupportedTypeError(EnvarifyError):
    """Unsupported type exception."""



"""Type inspection helper."""

import sys
from typing import Type, Union, get_args, get_origin

if sys.version_info >= (3, 10):
    from types import NoneType, UnionType
else:
    NoneType = type(None)
    UnionType = Union

from .errors import UnsupportedTypeError

SupportedType = Union[int, float, str, bool, dict, list, set, tuple]


class TypeInspector:
    """Helper class for type inspection and extraction."""

    def __init__(self, type_: Type, ignore_nullable: bool = False) -> None:
        """Initialize."""
        if ignore_nullable and (ti := TypeInspector(type_)):
            if ti.is_single_nullable():
                type_ = ti.extract_from_nullable()

        origin, args = get_origin(type_), get_args(type_)
        if origin is None:
            origin = type_

        self.type = origin[args] if args else origin
        self.origin_type = origin
        self.type_args = args

    def is_single_nullable(self) -> bool:
        """Check if type is a union of strictly one type and None."""
        is_union = self.origin_type is UnionType or self.origin_type is Union
        if not is_union:
            return False
        return len(self.type_args) == 2 and NoneType in self.type_args

    def is_dict(self) -> bool:
        """Check if type is a dictionary."""
        return self.origin_type is dict

    def is_sequence(self) -> bool:
        """Check if type is list. set or tuple."""
        return self.origin_type is list or self.origin_type is set or self.origin_type is tuple

    def extract_from_nullable(self) -> Type:
        """Extract base type from nullable/optional object."""
        if len(self.type_args) != 2:
            raise TypeError("Type {} is not nullable".format(self.type))

        if self.type_args[0] is NoneType:
            return self.type_args[1]
        elif self.type_args[1] is NoneType:
            return self.type_args[0]
        else:
            raise TypeError("Type {} is not nullable".format(self.type))

    def extract_inner_value_type(self) -> Type:
        """Extract inner value from type e.g. T from list[T]."""
        if len(self.type_args) > 1:
            raise UnsupportedTypeError(self.type)

        return self.type_args[0] if self.type_args else str



import typing as t
from unittest.mock import patch

import pytest

from envarify import EnvVar, UnsupportedTypeError, parse

from .const import PYTHON_IS_NEW


@pytest.mark.parametrize(
    "given, expected",
    [
        ("true", True),
        ("Yes", True),
        ("on", True),
        ("y", True),
        ("1", True),
        ("false", False),
        ("NO", False),
        ("OFF", False),
        ("n", False),
        ("0", False),
    ],
)
def test_str_to_bool_ok(given, expected):
    assert parse._str_to_bool(given) is expected


def test_str_to_bool_raises_error():
    with pytest.raises(ValueError):
        parse._str_to_bool("try")


def test_str_to_dict():
    sample_json = '{"A": 1, "B": "b", "C": {"D": true}}'
    expected = {"A": 1, "B": "b", "C": {"D": True}}
    assert parse._str_to_dict(sample_json) == expected


@pytest.mark.parametrize(
    "type_, expected",
    [
        (int, int),
        (t.Optional[int], int),
        (bool, parse._str_to_bool),
        (t.Union[str, None], str),
        (float, float),
        (dict, parse._str_to_dict),
    ]
    + (
        [
            (int | None, int),
            (None | str, str),
        ]
        if PYTHON_IS_NEW
        else []
    ),
)
def test_get_parser_ok(type_, expected):
    assert parse.get_parser(type_, EnvVar()) == expected


@patch("envarify.parse._get_sequence_parser", return_value="X")
def test_get_parser_returns_sequence_ok(mock_get_sequence_parser):
    assert parse.get_parser(list[int], EnvVar(delimiter="|")) == "X"
    mock_get_sequence_parser.assert_called_once_with(
        sequence_type=list,
        value_type=int,
        delimiter="|",
    )


def test_get_parser_raises_error():
    class SomeCringeType:
        pass

    with pytest.raises(UnsupportedTypeError):
        parse.get_parser(SomeCringeType, EnvVar())


@pytest.mark.parametrize(
    "outer, inner, value, expected",
    [
        (list, int, "1,2,3", [1, 2, 3]),
        (tuple, int, "1,2,3", (1, 2, 3)),
        (set, int, "1,2,3", {1, 2, 3}),
        (list, bool, "1,0,1", [True, False, True]),
    ],
)
def test_get_sequence_parser_ok(outer, inner, value, expected):
    assert parse._get_sequence_parser(outer, inner, ",")(value) == expected



import typing as t
from dataclasses import dataclass
from typing import Optional, Union
from unittest.mock import patch

import pytest

import envarify
from envarify import AnnotationError, BaseConfig, EnvVar, MissingEnvVarsError, UnsupportedTypeError
from envarify.envarify import _EnvVarSource

from .const import PYTHON_IS_NEW


class MyConfig(BaseConfig):
    x: int
    y: str


def test_base_config_init_ok():
    config = MyConfig(x=1, y="2")
    assert config.x == 1
    assert config.y == "2"


def test_base_config_init_raises_error():
    with pytest.raises(TypeError):
        MyConfig(z=1)


def test_base_config_repr_ok():
    assert MyConfig(x=1, y="2").__repr__() == "MyConfig(x=1, y='2')"


@patch.dict(
    envarify.envarify.os.environ,
    {
        "TEST_INT": "666",
        "TEST_FLOAT": "3.14",
        "TEST_STR": "Hello",
        "TEST_BOOL": "true",
        "TEST_SET": "1|2|3",
        "TEST_CUSTOM": "a,b,c",
    },
)
def test_base_config_fromenv_ok():

    class PrimitivesConfig(BaseConfig):
        test_int: int = EnvVar("TEST_INT")
        test_float: float = EnvVar("TEST_FLOAT")
        test_str: str = EnvVar("TEST_STR")
        test_bool: bool = EnvVar("TEST_BOOL")

    class MyConfig(BaseConfig):
        primitives: PrimitivesConfig
        test_set: t.Set[int] = EnvVar("TEST_SET", delimiter="|")
        test_custom: t.List[str] = EnvVar("TEST_CUSTOM", parse=lambda v: v.split(","))

    config = MyConfig.fromenv()
    assert config.primitives.test_int == 666
    assert config.primitives.test_float == 3.14
    assert config.primitives.test_str == "Hello"
    assert config.primitives.test_bool == True
    assert config.test_set == {1, 2, 3}
    assert config.test_custom == ["a", "b", "c"]


@patch.dict(envarify.envarify.os.environ, {"XXX": "666"})
def test_base_config_fromenv_nullable_arguments_ok():
    if PYTHON_IS_NEW:

        class MyConfig(BaseConfig):
            test_opt: t.Optional[int] = EnvVar("XXX")
            test_union: t.Union[float, None] = EnvVar("XXX")
            test_default: int | None = EnvVar(default=45)

    else:

        class MyConfig(BaseConfig):
            test_opt: t.Optional[int] = EnvVar("XXX")
            test_union: t.Union[float, None] = EnvVar("XXX")
            test_default: t.Optional[int] = EnvVar(default=45)

    config = MyConfig.fromenv()
    assert config.test_opt == 666
    assert config.test_union == 666.0
    assert config.test_default == 45


def test_base_config_fromenv_envvar_error_raised():
    class ConfigInside(BaseConfig):
        y: str = EnvVar("SOME_SUPER_UNLICKELY_ENVVAR")

    class MyConfig(BaseConfig):
        y: ConfigInside
        x: int = EnvVar("SOME_NOT_EXISTING_ENVVAR")

    with pytest.raises(MissingEnvVarsError) as e:
        MyConfig.fromenv()
        assert e.env_vars == ["SOME_NOT_EXISTING_ENVVAR", "SOME_SUPER_UNLICKELY_ENVVAR"]


def test_base_config_fromenv_unsupported_type_error_raised():
    class SomeCringeType:
        pass

    class MyConfig(BaseConfig):
        TEST: SomeCringeType

    with pytest.raises(UnsupportedTypeError):
        MyConfig.fromenv()


def test_base_config_fromenv_annotation_error_raised():
    class MyConfig(BaseConfig):
        pass

    with pytest.raises(AnnotationError):
        MyConfig.fromenv()


def test_base_config_envvars_ok():

    test_func = lambda x: x

    class MyConfig(BaseConfig):
        x: int = EnvVar("TEST_X", default=5)
        y: str = EnvVar(parse=test_func)

    assert MyConfig._envvars() == [
        _EnvVarSource(attr="x", name="TEST_X", default=5, parse=int),
        _EnvVarSource(attr="y", name="y", default=None, parse=test_func),
    ]


@patch.dict(envarify.envarify.os.environ, {"X": "25"})
def test_envvar_exists_ok():
    env_var = _EnvVarSource("x", "X", str)
    assert env_var.exists()


@patch.dict(envarify.envarify.os.environ, {})
def test_envvar_exists_not():
    env_var = _EnvVarSource("x", "X", str)
    assert not env_var.exists()


@patch("envarify.envarify._EnvVarSource.exists", return_value=True)
def test_envvar_has_value_ok(mock):
    env_var = _EnvVarSource("x", "X", str)
    assert env_var.has_value()


@patch("envarify.envarify._EnvVarSource.exists", return_value=False)
def test_envvar_has_value_ok_default(mock):
    env_var = _EnvVarSource("", "", str, default="x")
    assert env_var.has_value()


@patch("envarify.envarify._EnvVarSource.exists", return_value=False)
def test_envvar_has_value_ok_not(mock):
    env_var = _EnvVarSource("", "", str)
    assert not env_var.has_value()


def test_envvar_value():
    pass



import typing as t
from unittest.mock import patch

import pytest

from envarify import inspect

from .const import PYTHON_IS_NEW


def test_type_inspector_init_ok():
    ti = inspect.TypeInspector(t.Dict[str, int])
    assert ti.type == dict[str, int]
    assert ti.origin_type == dict
    assert ti.type_args == (str, int)


def test_type_inspector_init_ignore_nullable_ok():
    ti = inspect.TypeInspector(t.Optional[int], ignore_nullable=True)
    assert ti.type is int
    assert ti.origin_type is int
    assert ti.type_args is tuple()


@pytest.mark.parametrize(
    "type_, expected",
    (
        [
            (t.Optional[int], True),
            (t.Optional[t.List[int]], True),
            (t.Union[None, str], True),
            (t.Union[str, int, None], False),
            (int, False),
        ]
        + (
            [
                (int | None, True),
                (None | str, True),
                (str | int | None, False),
            ]
            if PYTHON_IS_NEW
            else []
        )
    ),
)
def test_type_inspector_is_single_nullable_ok(type_, expected):
    assert inspect.TypeInspector(type_).is_single_nullable() is expected


@pytest.mark.parametrize(
    "type_, expected",
    [
        (t.Dict, True),
        (dict, True),
        (set, False),
    ],
)
def test_type_inspector_is_dict_ok(type_, expected):
    assert inspect.TypeInspector(type_).is_dict() is expected


@pytest.mark.parametrize(
    "type_, expected",
    [
        (list, True),
        (t.List, True),
        (set, True),
        (tuple, True),
        (str, False),
    ],
)
def test_type_inspector_is_sequence_ok(type_, expected):
    assert inspect.TypeInspector(type_).is_sequence() is expected


@pytest.mark.parametrize(
    "type_, expected",
    [
        (t.Optional[int], int),
        (t.Union[None, str], str),
    ],
)
def test_type_inspector_extract_from_nullable_ok(type_, expected):
    assert inspect.TypeInspector(type_).extract_from_nullable() == expected


def test_type_inspector_extract_from_nullable_error():
    with pytest.raises(TypeError):
        inspect.TypeInspector(int).extract_from_nullable()

    with pytest.raises(TypeError):
        inspect.TypeInspector(int | str).extract_from_nullable()


@pytest.mark.parametrize(
    "type_, expected",
    [
        (list[int], int),
        (set[tuple[int]], tuple[int]),
    ],
)
def test_type_inspector_extract_inner_value_type_ok(type_, expected):
    assert inspect.TypeInspector(type_).extract_inner_value_type() == expected


def test_type_inspector_extract_inner_value_type_error():
    with pytest.raises(inspect.UnsupportedTypeError):
        inspect.TypeInspector(tuple[int, str]).extract_inner_value_type()



"""Handle module packaging."""

import re
import subprocess
import textwrap

from setuptools import setup

RE_VERSION = re.compile(r"(\d+\.\d+\.\d+)")


def _get_release_version() -> str:
    """Get git release tag version."""
    version_match = RE_VERSION.search(
        subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
        .stdout.decode("utf-8")
        .strip()
    )
    version = version_match[0] if version_match else "0.0.0"
    print("Release version: ", version)
    return version


def _write_version(version: str) -> None:
    """Write version to version.py."""
    with open("src/envarify/version.py", "w") as file:
        file.write(
            textwrap.dedent(
                f'''
                """Host package version, generated on build."""
                __version__ = "{version}"
                '''
            ).lstrip()
        )


version = _get_release_version()
_write_version(version)
setup(version=version)



"""Helpers for parsing from environment variable string to other types."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Callable, Type

from .errors import UnsupportedTypeError
from .inspect import SupportedType, TypeInspector

if TYPE_CHECKING:
    from .envarify import EnvVar


__all__ = ["get_parser", "EnvVarParser"]


EnvVarParser = Callable[[str], SupportedType]


def get_parser(type_: Type, spec: EnvVar) -> EnvVarParser:
    """Get parser for a given type."""
    ti = TypeInspector(type_, ignore_nullable=True)

    # primitive
    parser = _PRIMITIVE_PARSERS.get(ti.type)
    if parser is not None:
        return parser

    # dictionary
    if ti.is_dict():
        return _str_to_dict

    # sequence
    if ti.is_sequence():
        parser = _get_sequence_parser(
            sequence_type=ti.origin_type,
            value_type=ti.extract_inner_value_type(),
            delimiter=spec.delimiter,
        )
        if parser is not None:
            return parser

    raise UnsupportedTypeError(type_)


def _str_to_bool(value: str) -> bool:
    """Determine if string value is truthy."""
    value = value.lower()

    if value in _TRUE_VALUES:
        return True
    elif value in _FALSE_VALUES:
        return False

    raise ValueError("Cannot convert to bool: " + value)


def _str_to_dict(value: str) -> dict:
    """Convert string (JSON) to dictionary."""
    try:
        return json.loads(value)
    except json.decoder.JSONDecodeError:
        raise ValueError("Cannot convert to dictionary: " + value)


def _get_sequence_parser(sequence_type: Type, value_type: Type, delimiter: str) -> Callable | None:
    try:
        value_parser = _PRIMITIVE_PARSERS[value_type]
    except KeyError:
        return None

    def parser(sequence: str):
        return sequence_type(value_parser(value) for value in sequence.split(delimiter))

    return parser


_PRIMITIVE_PARSERS: dict[Type[SupportedType], EnvVarParser] = {
    int: int,
    float: float,
    str: str,
    bool: _str_to_bool,
}
_TRUE_VALUES = {"true", "yes", "on", "y", "1"}
_FALSE_VALUES = {"false", "no", "off", "n", "0"}



"""Envarify."""

from .envarify import BaseConfig, EnvVar
from .errors import AnnotationError, EnvarifyError, MissingEnvVarsError, UnsupportedTypeError

__all__ = [
    "BaseConfig",
    "EnvVar",
    "EnvarifyError",
    "AnnotationError",
    "MissingEnvVarsError",
    "UnsupportedTypeError",
]
